# -*- coding: utf8 -*-
# mbttsfactory.py
# 2010/08/16
# copyright(c) takayan
# http://neu101.seesaa.net/


__author__    = "takayan"
__copyright__ = "Copyright (c) 2010 takayan"
__license__   = "LGPL"

import mbtts

import re

import chnames
import debug
import speechserver
import settings
import orca
import orca_state
import punctuation_settings
from acss import ACSS
from orca_i18n import _

PUNCTUATION = re.compile('[^\w\s]', re.UNICODE)
ELLIPSIS = re.compile('(\342\200\246|\.\.\.\s*)')


class SpeechServer(speechserver.SpeechServer):
    # See the parent class for documentation.

    _default_voice_name = mbtts.default_database()
    mytts = mbtts.create(_default_voice_name)
    v_rate = 1.0
    v_volume = 1.0
    v_pitch  = 1.0
   
    _active_servers = {}
    
    DEFAULT_SERVER_ID = 'default'
    
    _SERVER_NAMES = {DEFAULT_SERVER_ID: _("Default Synthesizer")}
    
    KEY_NAMES = {
        '_':     'underscore',
        'space': 'space',
        '"':     'double-quote',
        }


    def getFactoryName():
        return _("mbtts")
    getFactoryName = staticmethod(getFactoryName)

    def getSpeechServers():
        servers = []
        default = SpeechServer._getSpeechServer(SpeechServer.DEFAULT_SERVER_ID)
        if default is not None:
            servers.append(default)
            for module in default.list_output_modules():
                servers.append(SpeechServer._getSpeechServer(module))
        return servers
    getSpeechServers = staticmethod(getSpeechServers)

    def _getSpeechServer(cls, serverId):
        if serverId not in cls._active_servers:
            cls(serverId)
        return cls._active_servers.get(serverId)
    _getSpeechServer = classmethod(_getSpeechServer)

    def getSpeechServer(info=None):
        if info is not None:
            thisId = info[1]
        else:
            thisId = SpeechServer.DEFAULT_SERVER_ID
        return SpeechServer._getSpeechServer(thisId)
    getSpeechServer = staticmethod(getSpeechServer)

    def shutdownActiveServers():
        for server in SpeechServer._active_servers.values():
            server.shutdown()
    shutdownActiveServers = staticmethod(shutdownActiveServers)

    # *** Instance methods ***

    def __init__(self, serverId):
        super(SpeechServer, self).__init__()
        self._id = serverId
        self._current_voice_properties = {}
        self._acss_manipulators = (
            (ACSS.RATE, self._set_rate),
            (ACSS.AVERAGE_PITCH, self._set_pitch),
            (ACSS.GAIN, self._set_volume),
            (ACSS.FAMILY, self._set_family),
            )
#        print "初期化実行"
        self._default_voice_name = mbtts.default_database()
        
        try:
            self._init()
        except:
            debug.println(debug.LEVEL_WARNING,
                          "mbtts failed to connect:")
            debug.printException(debug.LEVEL_WARNING)
        else:
            SpeechServer._active_servers[serverId] = self

    def _init(self):
        self._current_voice_properties = {}

    def _set_rate(self, acss_rate):
        rate = pow( 2.0, max( 0, min(99, acss_rate) ) * 0.06 - 3.0 )
        if ( rate < 0.2 ):
		rate = 0.2
#        print "rate:",rate
	self.v_rate = rate
        self.mytts.set_rate  ( self.v_rate   ) # 0.2 ... 8.0

    def _set_pitch(self, acss_pitch):
        pitch = pow( 2.0, max( 0, min(9, acss_pitch) ) * 0.6 - 3.0 )
        if ( pitch < 0.2 ):
		pitch = 0.2
#        print "pitch:",pitch
	self.v_pitch = pitch
        self.mytts.set_pitch ( self.v_pitch  ) # 0.2 ... 8.0

    def _set_volume(self, acss_volume):
        volume = max(0, min(9, acss_volume)) * 0.1
#        print "volume:",volume
	self.v_volume = volume
        self.mytts.set_volume( self.v_volume ) # 0.0 ... 1.0

    def _set_family(self, acss_family):
#        print "音声設定"
        name = acss_family[speechserver.VoiceFamily.NAME]
#        print "name:",name
	voices = mbtts.available_databases()
        if not name in voices:
        	name = self._default_voice_name
#		print "音声が見つからないので、標準音声に切り替えます"
        self.mytts = mbtts.create(name)
        self.mytts.set_rate  ( self.v_rate   ) # 0.2 ... 8.0
        self.mytts.set_pitch ( self.v_pitch  ) # 0.2 ... 8.0
        self.mytts.set_volume( self.v_volume ) # 0.0 ... 1.0
		
    def _apply_acss(self, acss):
        if acss is None:
            acss = settings.voices[settings.DEFAULT_VOICE]
        current = self._current_voice_properties
        for acss_property, method in self._acss_manipulators:
            value = acss.get(acss_property)
            if value is not None and current.get(acss_property) != value:
                method(value)
                current[acss_property] = value

    def __addVerbalizedPunctuation(self, oldText):
        spokenEllipsis = _(" dot dot dot")
        newText = re.sub(ELLIPSIS, spokenEllipsis, oldText)
        newText = newText.decode("UTF-8")

        symbols = set(re.findall(PUNCTUATION, newText))
        for symbol in symbols:
            try:
                level, action = punctuation_settings.getPunctuationInfo(symbol)
            except:
                continue

            if level != punctuation_settings.LEVEL_NONE:
                continue

            charName = " %s " % chnames.getCharacterName(symbol)
            if action == punctuation_settings.PUNCTUATION_INSERT:
                charName += symbol
            newText = re.sub(symbol, charName, newText)

        if orca_state.activeScript:
            newText = orca_state.activeScript.utilities.adjustForDigits(newText)

        return newText.encode("UTF-8")

    def _speak(self, text, acss, **kwargs):
        text = self.__addVerbalizedPunctuation(text)
        text = text.decode("UTF-8").replace(u'\u00a0', ' ').encode("UTF-8")
        text = text.replace('\n.', '\n')
        self._apply_acss(acss)
#        print text
        self.mytts.speak(unicode(text,"utf-8"),0,0)

    def _speak_char(self, text, acss, **kwargs):
        text = self.__addVerbalizedPunctuation(text)
        text = text.decode("UTF-8").replace(u'\u00a0', ' ').encode("UTF-8")
        text = text.replace('\n.', '\n')
        self._apply_acss(acss)
#        print text
        self.mytts.speak_spell(unicode(text,"utf-8"),0,0)

    def _cancel(self):
#        print "発音キャンセル"
        self.mytts.stop()

    def _change_default_speech_rate(self, decrease=False):
        acss = settings.voices[settings.DEFAULT_VOICE]
        delta = settings.speechRateDelta * (decrease and -1 or +1)
        rate = acss[ACSS.RATE]
        acss[ACSS.RATE] = max(0, min(99, rate + delta))
        debug.println(debug.LEVEL_CONFIGURATION,
                      "Speech rate is now %d" % rate)
        # Translators: This string announces speech rate change.
        self.speak(decrease and _("slower.") or _("faster."), acss=acss)

    def _change_default_speech_pitch(self, decrease=False):
        acss = settings.voices[settings.DEFAULT_VOICE]
        delta = settings.speechPitchDelta * (decrease and -1 or +1)
        pitch = acss[ACSS.AVERAGE_PITCH]
        acss[ACSS.AVERAGE_PITCH] = max(0, min(9, pitch + delta))
        debug.println(debug.LEVEL_CONFIGURATION,
                      "Speech pitch is now %d" % pitch)
        # Translators: This string announces speech pitch change.
        self.speak(decrease and _("lower.") or _("higher."), acss=acss)

    def getInfo(self):
        return [self._SERVER_NAMES.get(self._id, self._id), self._id]

    def getVoiceFamilies(self):
	databases = mbtts.available_databases()
	databases.sort()
	voices = []
	for database in databases:
		voices.append( (database,"ja","JP") )

        families = [speechserver.VoiceFamily({ \
              speechserver.VoiceFamily.NAME: name,
              speechserver.VoiceFamily.LOCALE: lang})
                    for name, lang, dialect in voices]
        return families

    def speak(self, text=None, acss=None, interrupt=True):
        if text:
            self._speak(text, acss)

    def queueText(self, text="", acss=None):
        if text:
            self._speak(text, acss)

    def speakUtterances(self, utteranceList, acss=None, interrupt=True):
        for utterance in utteranceList:
            if utterance:
                self._speak(utterance, acss)

    def speakCharacter(self, character, acss=None):
#        print "文字の読み上げ"
        self._apply_acss(acss)
        if character == '\n':
            self.speak('行末')
            return
#        print "文字名：",character
        self._speak_char(character, acss)
#        name = chnames.getCharacterName(character)
#        if not name:
#            print "文字：",character
#            self._speak_char(character, acss)
#            return
#
#        if orca_state.activeScript and orca_state.usePronunciationDictionary:
#            name = orca_state.activeScript.\
#                utilities.adjustForPronunciation(name)

    def speakKeyEvent(self, event_string, eventType):
        if eventType == orca.KeyEventType.PRINTABLE:
            if event_string.decode("UTF-8").isupper():
                acss = settings.voices[settings.UPPERCASE_VOICE]
            else:
                acss = None
            key = self.KEY_NAMES.get(event_string, event_string)
            self._speak(key, acss)
        else:
            return super(SpeechServer, self).speakKeyEvent(event_string, 
                                                           eventType)

    def increaseSpeechRate(self, step=5):
        self._change_default_speech_rate()

    def decreaseSpeechRate(self, step=5):
        self._change_default_speech_rate(decrease=True)

    def increaseSpeechPitch(self, step=0.5):
        self._change_default_speech_pitch()

    def decreaseSpeechPitch(self, step=0.5):
        self._change_default_speech_pitch(decrease=True)

    def stop(self):
        self._cancel()

    def shutdown(self):
        del SpeechServer._active_servers[self._id]

    def reset(self, text=None, acss=None):
        self._init()
        
    def list_output_modules(self):
        return ()

